import json
import logging
import os
import shutil
from functools import partial
import multiprocessing as mp
from typing import List, Tuple, Optional, TextIO
import time
import uuid

import xmltodict
from py7zr import py7zr

from .builder_helpers import extract_categories
from .logger_init import _init_logger_main_process, _init_logger_sub_process, _init_logger_multiprocessing
from .utils import get_file_list, prepare_output_dir, get_curr_version, cleanup_dir, compress_zstd, \
    COMPRESSION_EXTENSION
from .warehouse import Warehouse, get_warehouse_filenames

_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.INFO


class Builder:
    """
    Builder is a class for building the warehouse from the original data source.

    Attributes
    ----------
    output_dir : str
        The output directory.
    num_proc : int
        The number of processes to use.
    log_level : int
        The built-in logging level.
    compress : bool
        Whether to compress the output files.
    files : list
        A list of files to be read.

    """

    def __init__(self,
                 output_dir: str,
                 num_proc: int = _DEFAULT_NUM_PROC,
                 log_level: int = _DEFAULT_LOG_LEVEL,
                 max_size: int = 16,
                 compress: bool = True):
        """
        Parameters
        ----------
        output_dir : str
            The output directory.
        num_proc : int, default=1
            The number of processes to use.
        log_level : int, default=logging.INFO
            The built-in logging level.
        max_size : int, default=16
            The maximum size of an uncompressed warehouse in GB.
        compress : bool, default=True
            Whether to compress the output files.

        """
        self.output_dir = output_dir
        self.num_proc = num_proc
        self.log_level = log_level
        self.max_size = max_size
        self.compress = compress

        self.files: List[str] = []

        _init_logger_main_process(log_level=self.log_level)

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until the build() method is called.

        Parameters
        ----------
        path : str
            The path of a file or a directory.

        Raises
        ------
        ValueError
            If the path is empty.
        FileNotFoundError
            If the path does not exist.

        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def _worker_initializer(self, q):
        """
        Initialize the worker process.
        """
        # Initialize the logger within the sub-process.
        _init_logger_sub_process(q, log_level=self.log_level)

    def _decompress_executor(self, file_path: str) -> List[str]:
        # Initialize temporary directory.
        temp_dir = os.path.join(self.output_dir, 'temp')
        os.makedirs(temp_dir, exist_ok=True)

        archive_filename = os.path.basename(file_path)
        random_id = uuid.uuid4().hex
        decompressed_dir_path = os.path.join(temp_dir, random_id)
        os.makedirs(decompressed_dir_path, exist_ok=True)

        logging.debug(f'Decompressing [{archive_filename}]...')

        try:
            with py7zr.SevenZipFile(file_path, mode='r', mp=False) as z:
                start_time = time.time()
                z.extractall(path=decompressed_dir_path)
                end_time = time.time()
                execution_duration = (end_time - start_time) / 60
                logging.debug(f'Decompression took {execution_duration:.2f} min. ({archive_filename})')
            decompressed_files = get_file_list(decompressed_dir_path)
            return decompressed_files

        except Exception as e:
            logging.critical(f'Failed to decompress [{archive_filename}] for this reason: {e}')
            return []

    def _process_executor(self, xml_path: str, warehouse: Warehouse) -> List[str]:
        """
        Process the file.

        Parameters
        ----------
        xml_path : str
            The path of the file to be processed.
        warehouse : Warehouse
            The warehouse object.

        Returns
        -------
        List[str]
            The list of full paths of the warehouses.

        """
        logging.debug(f'Processing {os.path.basename(xml_path)}...')

        # Article Metadata.
        article_id: Optional[str] = None
        article_title: Optional[str] = None
        article_info: Optional[dict] = None

        assigned_warehouse: Optional[str] = None

        # Warehouse JSONL file path.
        warehouse_path: Optional[str] = None

        # Warehouse Metadata JSON file (which is metadata exclusively for this article).
        warehouse_metadata_path: Optional[str] = None

        # Warehouse JSONL IO.
        warehouse_file: Optional[TextIO] = None

        processed_count: int = 0
        full_warehouse_paths: List[str] = []

        def _create_new_article(new_article_id: str):
            nonlocal article_id, article_info, assigned_warehouse, warehouse_path, warehouse_metadata_path, \
                warehouse_file
            article_id = new_article_id
            # NOTE: We do not clean up the title because the title should be seen before the article id.
            article_info = dict()

            assigned_warehouse = warehouse.assign_warehouse()
            warehouse_filename, warehouse_metadata_filename = get_warehouse_filenames(assigned_warehouse)
            warehouse_path = os.path.join(self.output_dir, warehouse_filename)
            warehouse_metadata_path = os.path.join(self.output_dir, warehouse_metadata_filename)
            warehouse_file = open(warehouse_path, 'a')

            # Record the byte start of the article.
            article_info['byte_start'] = warehouse_file.tell()

        def _cleanup_article():
            nonlocal article_id, article_info, article_title, assigned_warehouse, warehouse_path, \
                warehouse_metadata_path, warehouse_file, full_warehouse_paths
            article_id = None
            article_info = None
            article_title = None
            warehouse_path = None
            warehouse_metadata_path = None
            if warehouse_file is not None:
                warehouse_file.close()
                warehouse_file = None
            if assigned_warehouse is not None:
                full_warehouse_path = warehouse.release_warehouse(assigned_warehouse)
                assigned_warehouse = None
                if full_warehouse_path is not None:
                    full_warehouse_paths.append(full_warehouse_path)

        def _finalizing_article():
            """
            After finishing writing an article, we will finalize the metadata file.
            """
            nonlocal article_id, article_info, article_title, assigned_warehouse, warehouse_path, \
                warehouse_metadata_path, warehouse_file, processed_count, full_warehouse_paths
            if article_id is None or warehouse_file is None:
                return
            with open(warehouse_metadata_path, 'a') as metadata_file:
                article_info['id'] = article_id
                article_info['title'] = article_title
                article_info['byte_end'] = warehouse_file.tell()
                if 'last_valid_text_content' in article_info:
                    article_info['categories'] = extract_categories(article_info['last_valid_text_content'])
                    del article_info['last_valid_text_content']
                else:
                    article_info['categories'] = []
                metadata_file.write(json.dumps(article_info) + '\n')
            warehouse_file.close()
            warehouse_file = None
            full_warehouse_path = warehouse.release_warehouse(assigned_warehouse)
            assigned_warehouse = None
            processed_count += 1
            if full_warehouse_path is not None:
                full_warehouse_paths.append(full_warehouse_path)

        def _inner_callback(path, item):
            nonlocal article_id, article_info, article_title, warehouse_file
            if path[-1][0] == 'title':
                # When a new article title is encountered, clean up the previous article and save the title for later.
                _finalizing_article()
                _cleanup_article()
                article_title = item.strip()
            if path[-1][0] == 'id':
                # When a new article id is encountered, create a new file for processing the article.
                _create_new_article(new_article_id=item.strip())
            if type(item) is dict and 'text' in item and '#text' in item['text']:
                to_be_written = dict()
                to_be_written['article_id'] = article_id
                if 'id' in item:
                    to_be_written['revision_id'] = item['id']
                    del item['id']
                if 'parentid' in item:
                    to_be_written['parent_id'] = item['parentid']
                    del item['parentid']
                if 'timestamp' in item:
                    to_be_written['timestamp'] = item['timestamp']
                    del item['timestamp']
                to_be_written.update(item)
                warehouse_file.write(json.dumps(to_be_written) + '\n')

                # Prepare text content for extracting categories.
                text_content = item['text']['#text']
                if len(text_content) > 0 and text_content[0] != '#REDIRECT':
                    # Extract categories.
                    article_info['source_revision'] = to_be_written['revision_id']
                    article_info['last_valid_text_content'] = text_content
            return True

        try:
            with open(xml_path, 'rb') as xml_file:
                xmltodict.parse(
                    xml_file,
                    item_depth=3,
                    item_callback=_inner_callback,
                )

            # Finalize the last article.
            _finalizing_article()

        except Exception as e:
            logging.critical(f'Error occurred when processing the file [{xml_file}]: {e}')

        try:
            # Remove XML file.
            os.remove(xml_path)

            # Remove the directory if it becomes empty.
            if not get_file_list(os.path.dirname(xml_path)):
                shutil.rmtree(os.path.dirname(xml_path))

        except:
            logging.error(f'Failed to remove the XML file [{xml_path}].')

        logging.debug(f'Processed (OK): count = {processed_count}')

        return full_warehouse_paths

    def _cleanup_executor(self, warehouse_filename: str):
        logging.debug(f'Warehouse cleaning: {warehouse_filename}')

        try:
            original_file_path = os.path.join(self.output_dir, warehouse_filename)
            if not os.path.exists(original_file_path):
                logging.critical(f'The file {warehouse_filename} [{original_file_path}] does not exist.')
                return None

            compressed_path = original_file_path + COMPRESSION_EXTENSION
            compress_zstd(original_file_path, compressed_path)
            os.remove(original_file_path)

            logging.debug(f'Warehouse cleaned (OK): {warehouse_filename}')

            return compressed_path

        except:
            logging.critical(f'Error occurred when compressing warehouse [{warehouse_filename}].')
            return None

    def build(self):
        """
        Build the blocks after apply the modifiers.
        """
        # Log the version.
        logging.info(f'Builder Version: {get_curr_version()}')

        warehouse = Warehouse(
            output_dir=self.output_dir,
            max_size=self.max_size,
            compress=self.compress,
        )

        start_time = time.time()

        # Prepare the output directory.
        prepare_output_dir(self.output_dir)

        total_count = len(self.files)

        global_temp_dir = os.path.join(self.output_dir, 'temp')
        os.makedirs(global_temp_dir, exist_ok=True)

        # Initialize multiprocessing logger.
        ql, q = _init_logger_multiprocessing(log_level=self.log_level)

        # Tasks is a list of tuples (task_type, args).
        # Task types: 'decompress', 'process', 'warehouse', 'cleanup'
        tasks: List[Tuple[str, Tuple]] = []

        available_process_count = self.num_proc

        modification_pool = mp.Pool(
            processes=self.num_proc,
            initializer=self._worker_initializer,
            initargs=(q,),
        )

        processed_count = 0

        def _decompress_callback(decompressed_files: List[str]):
            if not decompressed_files:
                logging.debug(f'Decompressed (EMPTY): {decompressed_files}')
                return

            decompressed_dir = os.path.dirname(decompressed_files[0])
            logging.debug(f'Decompressed (OK): {decompressed_dir} ({len(decompressed_files)})')

            for decompressed_file in decompressed_files:
                next_task_type = 'process'
                next_args = (decompressed_file,)
                tasks.insert(0, (next_task_type, next_args))

        def _process_callback(full_warehouse_paths: List[Tuple[str, str]]):
            nonlocal processed_count
            processed_count += 1

            logging.info(f'({processed_count / total_count * 100:.2f}% = {processed_count} / {total_count}) | '
                         f'Full warehouses: {len(full_warehouse_paths)}')

            for full_warehouse_path in full_warehouse_paths:
                next_task_type = 'cleanup'
                next_args = (full_warehouse_path,)
                tasks.insert(0, (next_task_type, next_args))

        def _cleanup_callback(cleanup_path):
            logging.info(f'Warehouse packed: {cleanup_path}')

        def _success_callback(task_type, file_path):
            nonlocal available_process_count
            if task_type == 'decompress':
                _decompress_callback(file_path)
            elif task_type == 'process':
                _process_callback(file_path)
            elif task_type == 'cleanup':
                _cleanup_callback(file_path)
            available_process_count += 1

        def _error_callback(e):
            nonlocal available_process_count
            logging.critical(f'Process terminated-level error: {e}')
            available_process_count += 1

        # Built-up initial tasks.
        for curr_file_path in self.files:
            tasks.append(('decompress', (curr_file_path,)))

        # Main semaphore loop.
        while tasks or available_process_count < self.num_proc:
            if tasks and available_process_count > 0:
                available_process_count -= 1
                task_type, args = tasks.pop(0)
                if task_type == 'decompress':
                    file_path, = args
                    modification_pool.apply_async(
                        func=self._decompress_executor,
                        args=(file_path,),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                elif task_type == 'process':
                    xml_file_path, = args
                    modification_pool.apply_async(
                        func=self._process_executor,
                        args=(xml_file_path, warehouse),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                elif task_type == 'cleanup':
                    warehouse_filename, = args
                    modification_pool.apply_async(
                        func=self._cleanup_executor,
                        args=(warehouse_filename,),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                else:
                    logging.critical(f'Unknown task type: {task_type}')
                    available_process_count += 1
            else:
                time.sleep(0.1)

        logging.info('Main loop finished. Waiting for cleanup...')

        if self.compress:
            for final_warehouse in warehouse.available_warehouses:
                next_task_type = 'cleanup'
                warehouse_filename = get_warehouse_filenames(final_warehouse)[0]
                next_args = (warehouse_filename,)
                tasks.append((next_task_type, next_args))

        # Final cleanup loop.
        while tasks or available_process_count < self.num_proc:
            if tasks and available_process_count > 0:
                available_process_count -= 1
                task_type, args = tasks.pop(0)
                if task_type == 'cleanup':
                    warehouse_filename, = args
                    modification_pool.apply_async(
                        func=self._cleanup_executor,
                        args=(warehouse_filename,),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                else:
                    logging.critical(f'Unknown task type: {task_type}')
                    available_process_count += 1
            else:
                time.sleep(0.1)

        logging.info(f'Cleanup loop finished.')

        # Clean up the global temporary directory.
        cleanup_dir(global_temp_dir)

        end_time = time.time()
        execution_duration = (end_time - start_time) / 60

        # Log the end of the task.
        logging.info(f'All done! Finished all files. (took {execution_duration:.2f} min in total)')



import logging
import os
from typing import Tuple, Union
import multiprocessing as mp


class Warehouse:
    """
    Warehouse is a module that manages the creation, assignment, and equal distribution of BloArk blocks (and segments).
    """

    def __init__(self,
                 output_dir: str,
                 prefix: str = 'warehouse_',
                 suffix: str = '',
                 max_size: int = 12,
                 compress: bool = True):
        self.output_dir = output_dir
        self.prefix = prefix
        self.suffix = suffix
        self.max_size = max_size
        self.compress = compress

        mp_manager = mp.Manager()
        self.mp_lock = mp_manager.Lock()
        self.warehouse_indexer = mp_manager.Value('i', 0)
        self.available_warehouses = mp_manager.list()
        self.occupied_warehouses = mp_manager.list()

    def create_warehouse(self):
        """
        Create a new warehouse when needed.
        This function will never return the created warehouse. Please try to access one via `assign_warehouse()`.

        Notes
        -----
        This function is intended to be called in main process (no parallelism).

        """
        # Fill 5 digits with 0s.
        curr_index = str(self.warehouse_indexer.value).zfill(5)
        new_filename_basename = f'{self.prefix}{curr_index}{self.suffix}'

        try:
            new_filename, new_metadata_filename = get_warehouse_filenames(new_filename_basename)
            new_filepath = os.path.join(self.output_dir, new_filename)
            new_metadata_filepath = os.path.join(self.output_dir, new_metadata_filename)
            with open(new_filepath, 'w') as f:
                f.truncate(0)
            with open(new_metadata_filepath, 'w') as f:
                f.truncate(0)
            self.available_warehouses.append(new_filename_basename)
            logging.debug(f'New warehouse created: {new_filename_basename}.')

        except:
            logging.error(f'Failed to create new warehouse: {new_filename_basename}.')

        self.warehouse_indexer.value += 1

    def assign_warehouse(self) -> str:
        """
        Request to assign a warehouse to a block.

        Returns
        -------
        assigned_warehouse : str
            The name of the assigned warehouse.

        Notes
        -----
        This function is intended to be called in main process (no parallelism).

        """
        with self.mp_lock:
            free_warehouses = [w for w in self.available_warehouses if w not in self.occupied_warehouses]
            if len(free_warehouses) == 0:
                self.create_warehouse()
                free_warehouses = [w for w in self.available_warehouses if w not in self.occupied_warehouses]

            # Find the warehouse with the smallest index. Assign the warehouse.
            assigned_warehouse = min(free_warehouses)
            self.occupied_warehouses.append(assigned_warehouse)

        return assigned_warehouse

    def release_warehouse(self, warehouse: str) -> Union[str, None]:
        """
        Release the assignment of a warehouse.

        Parameters
        ----------
        warehouse : str
            The name of the warehouse to be released.

        Returns
        -------
        warehouse_file_should_compress : str or None
            The name of the warehouse file that should be compressed. If None, no compression is needed.

        Notes
        -----
        This function is intended to be called in main process (no parallelism).

        """
        warehouse_file_should_compress = None

        with self.mp_lock:
            try:
                self.occupied_warehouses.remove(warehouse)

                # Check current size, if it is larger than the max size, remove it from the available warehouses.
                warehouse_file, warehouse_metadata_file = get_warehouse_filenames(warehouse)
                warehouse_file_size = get_file_size(os.path.join(self.output_dir, warehouse_file))
                if warehouse_file_size >= self.max_size:
                    self.available_warehouses.remove(warehouse)
                    if self.compress:
                        warehouse_file_should_compress = warehouse_file

            except:
                logging.error(f'Warehouse [{warehouse}] does not exist.')

        return warehouse_file_should_compress

    def finalize_warehouse(self, warehouse: str):
        """
        Finalize a warehouse.
        This function should be called when the warehouse is full and no more blocks will be assigned to it.

        Parameters
        ----------
        warehouse : str
            The name of the warehouse to be finalized.

        Notes
        -----
        This function is intended to be called in main process (no parallelism).

        """
        with self.mp_lock:
            self.available_warehouses.remove(warehouse)


def get_warehouse_filenames(basename: str) -> Tuple[str, str]:
    filename = f'{basename}.jsonl'
    metadata_filename = f'{basename}.metadata'
    return filename, metadata_filename


def get_file_size(filepath: str) -> float:
    if not os.path.exists(filepath):
        return 0
    return os.path.getsize(filepath) / (1000 ** 3)



import json
import logging
import os
from functools import partial
import multiprocessing as mp
from typing import List, Tuple, Union
import time
import uuid

from py7zr import py7zr

from .logger_init import _init_logger_main_process, _init_logger_sub_process, _init_logger_multiprocessing
from .utils import get_file_list, prepare_output_dir, get_curr_version, cleanup_dir, read_line_in_file, parse_schema

_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.INFO


class Reader:
    """
    Reader is a class for reading the data from the warehouse (rather than from the original data source).

    Attributes
    ----------
    output_dir : str
        The output directory.
    num_proc : int
        The number of processes to use.
    log_level : int
        The built-in logging level.
    files : list
        A list of files to be read.

    """

    def __init__(self,
                 output_dir: str,
                 num_proc: int = _DEFAULT_NUM_PROC,
                 log_level: int = _DEFAULT_LOG_LEVEL):
        """
        Parameters
        ----------
        output_dir : str
            The output directory.
        num_proc : int, default=1
            The number of processes to use.
        log_level : int, default=logging.INFO
            The built-in logging level.

        """
        self.output_dir = output_dir
        self.num_proc = num_proc
        self.log_level = log_level

        self.files: List[str] = []

        _init_logger_main_process(log_level=self.log_level)

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until any other method is called.

        Parameters
        ----------
        path : str
            The path of a file or a directory.

        Raises
        ------
        ValueError
            If the path is empty.
        FileNotFoundError
            If the path does not exist.

        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def _worker_initializer(self, q):
        """
        Initialize the worker process.
        """
        # Initialize the logger within the sub-process.
        _init_logger_sub_process(q, log_level=self.log_level)

    def _decompress_executor(self, file_path: str) -> List[str]:
        """
        Decompress the file.
        """
        # Initialize temporary directory.
        temp_dir = os.path.join(self.output_dir, 'temp')
        os.makedirs(temp_dir, exist_ok=True)

        archive_filename = os.path.basename(file_path)
        random_id = uuid.uuid4().hex
        decompressed_dir_path = os.path.join(temp_dir, random_id)
        os.makedirs(decompressed_dir_path, exist_ok=True)

        logging.debug(f'Decompressing [{archive_filename}]...')

        try:
            with py7zr.SevenZipFile(file_path, mode='r', mp=False) as z:
                start_time = time.time()
                z.extractall(path=decompressed_dir_path)
                end_time = time.time()
                execution_duration = (end_time - start_time) / 60
                logging.debug(f'Decompression took {execution_duration:.2f} min. ({archive_filename})')
            decompressed_files = get_file_list(decompressed_dir_path)
            return decompressed_files

        except Exception as e:
            logging.critical(f'Failed to decompress [{archive_filename}] for this reason: {e}')
            return []

    def decompress(self):
        """
        Decompress the preloaded files.
        """
        # Log the version.
        logging.info(f'Builder Version: {get_curr_version()}')

        start_time = time.time()

        # Prepare the output directory.
        prepare_output_dir(self.output_dir)

        total_count = len(self.files)

        # Initialize multiprocessing logger.
        ql, q = _init_logger_multiprocessing(log_level=self.log_level)

        # Tasks is a list of tuples (task_type, args).
        # Task types: 'decompress', 'process', 'warehouse', 'cleanup'
        tasks: List[Tuple[str, Tuple]] = []

        available_process_count = self.num_proc

        decompression_pool = mp.Pool(
            processes=self.num_proc,
            initializer=self._worker_initializer,
            initargs=(q,),
        )

        processed_count = 0

        def _decompress_callback(decompressed_files: List[str]):
            nonlocal processed_count, total_count
            if not decompressed_files:
                logging.debug(f'Decompressed (EMPTY): {decompressed_files}')
                return

            processed_count += 1

            decompressed_dir = os.path.dirname(decompressed_files[0])
            logging.debug(f'Decompressed (OK): {decompressed_dir} ({processed_count} / {total_count})')

        def _success_callback(task_type, file_path):
            nonlocal available_process_count
            if task_type == 'decompress':
                _decompress_callback(file_path)
            available_process_count += 1

        def _error_callback(e):
            nonlocal available_process_count
            logging.critical(f'Process terminated-level error: {e}')
            available_process_count += 1

        # Built-up initial tasks.
        for curr_file_path in self.files:
            tasks.append(('decompress', (curr_file_path,)))

        # Main semaphore loop.
        while tasks or available_process_count < self.num_proc:
            if tasks and available_process_count > 0:
                available_process_count -= 1
                task_type, args = tasks.pop(0)
                if task_type == 'decompress':
                    file_path, = args
                    decompression_pool.apply_async(
                        func=self._decompress_executor,
                        args=(file_path,),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                else:
                    logging.critical(f'Unknown task type: {task_type}')
                    available_process_count += 1
            else:
                time.sleep(0.1)

        logging.info('Main loop finished.')

        end_time = time.time()
        execution_duration = (end_time - start_time) / 60

        # Log the end of the task.
        logging.info(f'All done! Finished {processed_count} files. (took {execution_duration:.2f} min in total)')

    def glimpse(self) -> Union[Tuple[dict, dict], Tuple[None, None]]:
        """
        Take a glimpse of the preloaded data.
        It could still be large if one object contains a lot of information (e.g. many revisions, long article).

        Returns
        -------
        Tuple[dict, dict]
            A tuple of two dictionaries: (page, revision). If there is no file loaded, it returns (None, None).

        Notes
        -----
        This function does not use any parallelization technique.

        """
        if len(self.files) == 0:
            logging.warning('No file is loaded.')
            return None, None

        # Randomly select a file.
        picked_index = int(uuid.uuid4().int % len(self.files))
        glimpse_path = self.files[picked_index]

        logging.info(f'Randomly chosen file: {glimpse_path}')

        # Prepare the temporary directory for glimpse.
        glimpse_temp_dir = os.path.join(os.getcwd(), '.glimpse')
        os.makedirs(glimpse_temp_dir, exist_ok=True)

        logging.info(f'Decompressing...')

        # Decompress the file.
        decompressed_files = self._decompress_executor(glimpse_path)
        if not decompressed_files:
            logging.error(f'Failed to decompress the file: {glimpse_path}')
            return None, None

        decompressed_path = decompressed_files[0]
        first_block_text = read_line_in_file(decompressed_path, 0).rstrip('\n')

        if first_block_text[0] != '{' or first_block_text[-1] != '}':
            logging.error(f'Invalid starting of block or end of block.')
            return None, None

        # Read the first block into memory and then delete the decompressed file.
        first_block = json.loads(first_block_text)
        cleanup_dir(glimpse_temp_dir)

        logging.info(f'Glimpse finished.')

        return first_block, parse_schema(first_block)



import logging
import multiprocessing as mp
from logging.handlers import QueueListener, QueueHandler

_log_format = "[%(asctime)s (%(process)d) %(levelname)s] %(message)s"
_formatter = logging.Formatter(_log_format)

_main_log_format = "[%(asctime)s (main) %(levelname)s] %(message)s"
_main_formatter = logging.Formatter(_main_log_format)


def _get_logger_stream_handler(main_process: bool, log_level: int) -> logging.StreamHandler:
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(_formatter if not main_process else _main_formatter)
    stream_handler.setLevel(log_level)
    return stream_handler


def _cleanup_logger_handlers(logger: logging.Logger):
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])


def _init_logger_main_process(log_level: int):
    stream_handler = _get_logger_stream_handler(main_process=True, log_level=log_level)
    logger = logging.getLogger()
    _cleanup_logger_handlers(logger)
    logger.setLevel(log_level)
    logger.addHandler(stream_handler)


def _init_logger_multiprocessing(log_level: int = logging.INFO) -> (QueueListener, mp.Queue):
    q = mp.Queue()

    stream_handler = _get_logger_stream_handler(main_process=False, log_level=log_level)

    ql = QueueListener(q, stream_handler)
    ql.start()

    # I use a different name here to avoid conflict with the default logger.
    logger = logging.getLogger('mp_parent_logger')
    logger.setLevel(log_level)
    logger.addHandler(stream_handler)

    return ql, q


def _init_logger_sub_process(q: mp.Queue, log_level: int = logging.INFO):
    qh = QueueHandler(q)
    logger = logging.getLogger()

    # Clean up any existing handlers in subprocess default logger.
    _cleanup_logger_handlers(logger)
    logger.setLevel(log_level)
    logger.addHandler(qh)



import json
import logging
import os
from functools import partial
from multiprocessing import Pool
from typing import List, Tuple, Optional
from abc import ABC, abstractmethod
import time

from .logger_init import _init_logger_main_process, _init_logger_sub_process, _init_logger_multiprocessing
from .utils import get_file_list, decompress_zstd, prepare_output_dir, get_curr_version, \
    cleanup_dir, compress_zstd, COMPRESSION_EXTENSION, unstable
from .warehouse import Warehouse, get_warehouse_filenames

_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.INFO


class ModifierProfile(ABC):
    """
    The core class to define how to modify the JSON content.
    """
    @abstractmethod
    def block(self, content: dict, metadata: dict) -> Optional[dict]:
        """
        Returns a list of batches of URLs to download.

        Parameters
        ----------
        content : dict
            The JSON content to be modified.
        metadata : dict
            The metadata of the JSON content. This will be updated within one segment from the previous return value.

        Returns
        -------
        Tuple[Optional[dict], Optional[dict]]
            Data JSON content and metadata (whatever modified or not).
            Return None in the first value if the content should be removed.
            Return None in the second value if the entire segment should be removed.

        """
        pass


@unstable
class Modifier:
    """
    Modifier is the class to define how to modify the JSON content of a block (or a segment) from the warehouse.

    Attributes
    ----------
    num_proc : int
        The number of processes to use.
    log_level : int
        The log level.
    files : list
        A list of files to be read.
    modifiers : list
        A list of modifiers to be applied.

    """

    def __init__(self,
                 output_dir: str,
                 num_proc: int = _DEFAULT_NUM_PROC,
                 log_level: int = _DEFAULT_LOG_LEVEL):
        """
        Parameters
        ----------
        output_dir : str
            The output directory.
        num_proc : int, default=1
            The number of processes to use.
        log_level : int, default=logging.INFO
            The log level.

        """
        self.output_dir = output_dir
        self.num_proc = num_proc
        self.log_level = log_level

        self.files: List[str] = []
        self.modifiers: List[ModifierProfile] = []

        _init_logger_main_process(log_level=self.log_level)

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until the build() method is called.

        Parameters
        ----------
        path : str
            The path of a file or a directory.

        Raises
        ------
        ValueError
            If the path is empty.
        FileNotFoundError
            If the path does not exist.

        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def add_profile(self, profile: ModifierProfile):
        """
        Map a function to each block.

        Parameters
        ----------
        profile : ModifierProfile
            The modifier profile to be added.

        """
        self.modifiers.append(profile)

    def _worker_initializer(self, q):
        """
        Initialize the worker process.
        """
        # Initialize the logger within the sub-process.
        _init_logger_sub_process(q, log_level=self.log_level)

    def _process_executor(self, file_path: str):
        temp_dir = os.path.join(self.output_dir, 'temp')
        os.makedirs(temp_dir, exist_ok=True)

        original_name = os.path.split(file_path)[1]
        decompressed_name = original_name.replace('.zst', '')
        decompressed_path = os.path.join(temp_dir, decompressed_name)

        # Decompress the file.
        decompress_zstd(file_path, decompressed_path)

        modified_name = decompressed_name + '.modified'
        modified_path = os.path.join(temp_dir, modified_name)

        modified_file = open(modified_path, 'w')

        # TODO: Get metadata from the corresponding warehouse.

        # Read the file line by line, and apply the modifiers.
        with open(decompressed_path, 'r') as f:
            for line in iter(f.readline, ''):
                block = json.loads(line)
                for modifier in self.modifiers:
                    block = modifier.block(block)
                    if block is None:
                        break
                if block is not None:
                    modified_file.write(json.dumps(block) + '\n')

        modified_file.close()

        # Remove the decompressed file.
        os.remove(decompressed_path)

        return modified_path

    def _warehouse_executor(self, file_path: str, assigned_warehouse: str):
        try:
            warehouse_filename, warehouse_metadata_filename = get_warehouse_filenames(assigned_warehouse)
            warehouse_path = os.path.join(self.output_dir, warehouse_filename)
            # warehouse_metadata_path = os.path.join(self.output_dir, warehouse_metadata_filename)  # TODO.

            warehouse_file = open(warehouse_path, 'a')

            # Read the file line by line, and move directly to the warehouse (without last empty line).
            with open(file_path, 'r') as f:
                for line in iter(f.readline, ''):
                    warehouse_file.write(line + '\n')

            warehouse_file.close()

            # Remove the modified file.
            os.remove(file_path)

        except Exception as e:
            logging.critical(f'Error occurred when moving the file to the warehouse. {e}')

        return assigned_warehouse

    def _cleanup_executor(self, warehouse_path: str):
        try:
            original_file_path = os.path.join(self.output_dir, warehouse_path)
            if not os.path.exists(original_file_path):
                logging.critical(f'The file {warehouse_path} [{original_file_path}] does not exist.')
                return None

            compressed_path = original_file_path + COMPRESSION_EXTENSION
            compress_zstd(original_file_path, compressed_path)
            os.remove(original_file_path)

            return compressed_path

        except:
            logging.critical(f'Error occurred when compressing the file {warehouse_path}.')
            return None

    def build(self):
        """
        Build the blocks after apply the modifiers.
        """
        # Log the version.
        logging.info(f'Modifier Version: {get_curr_version()}')

        warehouse = Warehouse(
            output_dir=self.output_dir,
            max_size=8,
            compress=True,
        )

        start_time = time.time()

        # Prepare the output directory.
        prepare_output_dir(self.output_dir)

        curr_count = 0
        total_count = len(self.files)

        global_temp_dir = os.path.join(self.output_dir, 'temp')
        os.makedirs(global_temp_dir, exist_ok=True)

        # Initialize multiprocessing logger.
        ql, q = _init_logger_multiprocessing(log_level=self.log_level)

        # Tasks is a list of tuples (task_type, args).
        # Task types: 'process', 'warehouse', 'cleanup'
        tasks: List[Tuple[str, Tuple]] = []

        available_process_count = self.num_proc

        modification_pool = Pool(
            processes=self.num_proc,
            initializer=self._worker_initializer,
            initargs=(q,)
        )

        processed_count = 0
        finished_count = 0

        def _process_callback(file_path):
            nonlocal warehouse, available_process_count, processed_count
            processed_count += 1

            assigned_warehouse = warehouse.assign_warehouse()

            logging.info(f'({processed_count / total_count * 100:.2f}% = {processed_count} / {total_count}) | '
                         f'Processed: {os.path.basename(file_path)} | Warehouse: {assigned_warehouse}')

            next_task_type = 'warehouse'
            next_args = (file_path, assigned_warehouse)
            tasks.insert(0, (next_task_type, next_args))

            available_process_count += 1

        def _warehouse_callback(warehouse_basename):
            nonlocal warehouse, available_process_count, finished_count

            logging.debug(f'Saved into warehouse: {warehouse_basename}')

            cleanup_path = warehouse.release_warehouse(warehouse_basename)
            if cleanup_path is not None:
                next_task_type = 'cleanup'
                next_args = (cleanup_path,)
                tasks.insert(0, (next_task_type, next_args))

            finished_count += 1
            available_process_count += 1

        def _cleanup_callback(cleanup_path):
            nonlocal available_process_count
            logging.info(f'Warehouse packed: {cleanup_path}')

            available_process_count += 1

        def _success_callback(task_type, file_path):
            if task_type == 'process':
                _process_callback(file_path)
            elif task_type == 'warehouse':
                _warehouse_callback(file_path)
            elif task_type == 'cleanup':
                _cleanup_callback(file_path)

        def _error_callback(e):
            logging.error(f'Error occurred when processing block: {e}')

        # Built-up initial tasks.
        for curr_file_path in self.files:
            tasks.append(('process', (curr_file_path,)))

        # Main semaphore loop.
        while tasks or available_process_count < self.num_proc:
            if tasks and available_process_count > 0:
                available_process_count -= 1
                task_type, args = tasks.pop(0)
                if task_type == 'process':
                    file_path, = args
                    modification_pool.apply_async(
                        func=self._process_executor,
                        args=(file_path,),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                elif task_type == 'warehouse':
                    file_path, assigned_warehouse = args
                    modification_pool.apply_async(
                        func=self._warehouse_executor,
                        args=(file_path, assigned_warehouse),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
                elif task_type == 'cleanup':
                    cleanup_path, = args
                    modification_pool.apply_async(
                        func=self._cleanup_executor,
                        args=(cleanup_path,),
                        callback=partial(_success_callback, task_type),
                        error_callback=_error_callback
                    )
            else:
                time.sleep(0.1)

        logging.debug(f'Semaphore loop finished.')

        # Clean up the global temporary directory.
        cleanup_dir(global_temp_dir)

        end_time = time.time()
        execution_duration = end_time - start_time

        # Log the end of the task.
        logging.info(f'All done! Finished all files. (took {execution_duration:.2f}s in total)')
